package org.dotcms.plugins.htmlPageExtension;

import java.io.File;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

import com.dotmarketing.beans.Host;
import com.dotmarketing.beans.Identifier;
import com.dotmarketing.business.APILocator;
import com.dotmarketing.cache.FieldsCache;
import com.dotmarketing.cache.LiveCache;
import com.dotmarketing.cache.VirtualLinksCache;
import com.dotmarketing.common.db.DotConnect;
import com.dotmarketing.db.DbConnectionFactory;
import com.dotmarketing.db.HibernateUtil;
import com.dotmarketing.exception.DotDataException;
import com.dotmarketing.exception.DotHibernateException;
import com.dotmarketing.factories.InodeFactory;
import com.dotmarketing.factories.WebAssetFactory;
import com.dotmarketing.plugin.PluginDeployer;
import com.dotmarketing.plugin.business.PluginAPI;
import com.dotmarketing.portlets.cmsmaintenance.action.ViewCMSMaintenanceAction;
import com.dotmarketing.portlets.containers.factories.ContainerFactory;
import com.dotmarketing.portlets.containers.model.Container;
import com.dotmarketing.portlets.contentlet.business.ContentletAPI;
import com.dotmarketing.portlets.contentlet.model.Contentlet;
import com.dotmarketing.portlets.folders.business.FolderAPI;
import com.dotmarketing.portlets.folders.model.Folder;
import com.dotmarketing.portlets.htmlpages.factories.HTMLPageFactory;
import com.dotmarketing.portlets.htmlpages.model.HTMLPage;
import com.dotmarketing.portlets.links.factories.LinkFactory;
import com.dotmarketing.portlets.links.model.Link;
import com.dotmarketing.portlets.structure.business.FieldAPI;
import com.dotmarketing.portlets.structure.factories.FieldFactory;
import com.dotmarketing.portlets.structure.factories.StructureFactory;
import com.dotmarketing.portlets.structure.model.Field;
import com.dotmarketing.portlets.structure.model.Structure;
import com.dotmarketing.portlets.templates.factories.TemplateFactory;
import com.dotmarketing.portlets.templates.model.Template;
import com.dotmarketing.portlets.virtuallinks.factories.VirtualLinkFactory;
import com.dotmarketing.portlets.virtuallinks.model.VirtualLink;
import com.dotmarketing.util.Config;
import com.dotmarketing.util.Logger;
import com.dotmarketing.util.MaintenanceUtil;
import com.dotmarketing.util.UtilMethods;
import com.liferay.portal.model.User;

/**
 * This plugin deployer change page extensions
 * @author Armando, Oswaldo
 *
 */
public class HtmlPageExtensionPluginDeployer implements PluginDeployer {

	private PluginAPI pluginApi = APILocator.getPluginAPI();
	private ContentletAPI conAPI = APILocator.getContentletAPI();
	private FolderAPI folderAPI = APILocator.getFolderAPI();
	private String oldExt;
	private String newExt;
	final String PGUPDATESQL1 = "UPDATE HTMLPAGE SET PAGE_URL = REPLACE(PAGE_URL, ?, ?) WHERE PAGE_URL LIKE ?";
	final String PGUPDATESQL2 = "UPDATE IDENTIFIER SET ASSET_NAME = REPLACE(ASSET_NAME, ?, ?) WHERE ASSET_NAME LIKE ?";
	final String MYUPDATESQL1 = "UPDATE HTMLPAGE SET PAGE_URL = REPLACE(PAGE_URL, ?, ?) WHERE PAGE_URL LIKE ?";
	final String MYUPDATESQL2 = "UPDATE IDENTIFIER SET ASSET_NAME = REPLACE(ASSET_NAME, ?, ?) WHERE ASSET_NAME LIKE ?";
	final String MSUPDATESQL1 = "UPDATE HTMLPAGE SET PAGE_URL = REPLACE(PAGE_URL, ?, ?) WHERE PAGE_URL LIKE ?";
	final String MSUPDATESQL2 = "UPDATE IDENTIFIER SET ASSET_NAME = REPLACE(URI, ?, ?) WHERE ASSET_NAME LIKE ?";
	final String OCLUPDATESQL1 = "UPDATE HTMLPAGE SET PAGE_URL = REPLACE(PAGE_URL, ?, ?) WHERE PAGE_URL LIKE ?";
	final String OCLUPDATESQL2 = "UPDATE IDENTIFIER SET ASSET_NAME = REPLACE(URI, ?, ?) WHERE ASSET_NAME LIKE ?";



	/**
	 * This method run the routine to change dotCMS site page extensions
	 */
	public boolean deploy() {

		try{
			oldExt = pluginApi.loadProperty("org.dotcms.plugins.htmlPageExtension", "OLD_HTMLPAGE_EXTENSION");
			newExt = pluginApi.loadProperty("org.dotcms.plugins.htmlPageExtension", "NEW_HTMLPAGE_EXTENSION");

			searchAndReplace("\\." + oldExt, "." + newExt);

			DotConnect dc = new DotConnect();
			if(DbConnectionFactory.getDBType().equals(DbConnectionFactory.POSTGRESQL)){
				dc.setSQL(PGUPDATESQL1);
				dc.addParam("." + oldExt);
				dc.addParam("." + newExt);
				dc.addParam("%." + oldExt);
				try {
					dc.loadResult();
				} catch (DotDataException e) {
					Logger.error(this, e.getMessage(), e);
				}
				dc.setSQL(PGUPDATESQL2);
				dc.addParam("." + oldExt);
				dc.addParam("." + newExt);
				dc.addParam("%." + oldExt);
				try {
					dc.loadResult();
				} catch (DotDataException e) {
					Logger.error(this, e.getMessage(), e);
				}
			}else if(DbConnectionFactory.getDBType().equals(DbConnectionFactory.MYSQL)){
				dc.setSQL(MYUPDATESQL1);
				dc.addParam("." + oldExt);
				dc.addParam("." + newExt);
				dc.addParam("%." + oldExt);
				try {
					dc.loadResult();
				} catch (DotDataException e) {
					Logger.error(this, e.getMessage(), e);
				}
				dc.setSQL(MYUPDATESQL2);
				dc.addParam("." + oldExt);
				dc.addParam("." + newExt);
				dc.addParam("%." + oldExt);
				try {
					dc.loadResult();
				} catch (DotDataException e) {
					Logger.error(this, e.getMessage(), e);
				}
			}else if(DbConnectionFactory.getDBType().equals(DbConnectionFactory.MSSQL)){
				dc.setSQL(MSUPDATESQL1);
				dc.addParam("." + oldExt);
				dc.addParam("." + newExt);
				dc.addParam("%." + oldExt);
				try {
					dc.loadResult();
				} catch (DotDataException e) {
					Logger.error(this, e.getMessage(), e);
				}
				dc.setSQL(MSUPDATESQL2);
				dc.addParam("." + oldExt);
				dc.addParam("." + newExt);
				dc.addParam("%." + oldExt);
				try {
					dc.loadResult();
				} catch (DotDataException e) {
					Logger.error(this, e.getMessage(), e);
				}
			}else{
				dc.setSQL(OCLUPDATESQL1);
				dc.addParam("." + oldExt);
				dc.addParam("." + newExt);
				dc.addParam("%." + oldExt);
				try {
					dc.loadResult();
				} catch (DotDataException e) {
					Logger.error(this, e.getMessage(), e);
				}
				dc.setSQL(OCLUPDATESQL2);
				dc.addParam("." + oldExt);
				dc.addParam("." + newExt);
				dc.addParam("%." + oldExt);
				try {
					dc.loadResult();
				} catch (DotDataException e) {
					Logger.error(this, e.getMessage(), e);
				}
			}
		}catch(Exception e){
			Logger.error(HtmlPageExtensionPluginDeployer.class, e.getMessage(), e);
		}finally{
			DbConnectionFactory.closeConnection();
		}

		return true;
	}

	/**
	 * Runs deploy process when the manifest version is updated
	 */
	public boolean redeploy(String version) {
		return deploy();
	}

	/**
	 * This method update the contentlets, links, structures, templates, containers and pages links
	 * with the new page extension
	 * @param searchString
	 * @param replaceString
	 */
	private void searchAndReplace(String searchString, String replaceString) {

		User systemUser = null;

		Map<String,Object> params = new HashMap<String, Object>();

		Host host = null;

		if (UtilMethods.isSet(searchString) && UtilMethods.isSet(replaceString)) {
			try {
				HibernateUtil.startTransaction();
			} catch (DotHibernateException e2) {
				Logger.warn(this,"Unable to start transaction : " + e2.getMessage(), e2);
				Logger.error(this,"Unable to start transaction : " + e2.getMessage(), e2);
			}

			try{

				host = APILocator.getHostAPI().findDefaultHost(systemUser, true);

				systemUser = APILocator.getUserAPI().getSystemUser();

				Logger.info(this, "Searching for Contentlets");
				String originalPatter = searchString;
				String newPatter = replaceString;
				Pattern patter = Pattern.compile(originalPatter);
				Matcher matcher = patter.matcher("");
				int cicleValue = 1000;
				int offSet = 0;
				int contentletsCounter = 0;
				int contentletsProccesed = 0;

				List<Contentlet> contentlets = conAPI.search("+working:true -structureName:Host", cicleValue, offSet, "", APILocator.getUserAPI().getSystemUser(), false);
				int size = contentlets.size();
				Logger.info(this, "Processing first set of " + size);
				while(size > 0)
				{
					for(int i = 0;i < size;i++)
					{
						Contentlet contentlet = contentlets.get(i);

						try {
							boolean changed = false;
							Logger.debug(this, "Modifying Contentlet with inode " + contentlet.getInode());
							List<Field> fields = FieldsCache.getFieldsByStructureInode(contentlet.getStructureInode());
							for (Field field : fields) {
								FieldAPI fapi = APILocator.getFieldAPI();
								boolean stringProperty = fapi.isString(field);
								if(stringProperty){
									String value = "";
									Object fValue = conAPI.getFieldValue(contentlet, field);
									if(fValue == null){
										fValue = "";
									}else{
										if(fValue instanceof String)
											value = (String)fValue;
									}
									if(UtilMethods.isSet(value)){
										matcher.reset(value);
										if(matcher.find()){
											changed = true;
											conAPI.setContentletProperty(contentlet, field, matcher.replaceAll(newPatter));
										}
									}
									else {
										changed = false;
									}
								}
							}

							if(changed)
							{
								//contentlet.setInode(null);
								contentlet.setInode("");
								contentlet = conAPI.checkin(contentlet,APILocator.getUserAPI().getSystemUser(), true);
								APILocator.getVersionableAPI().setWorking(contentlet);
								contentletsCounter++;
							}
							contentletsProccesed++;
						} catch (Exception e) {
							Logger.error(this, "Unable to checkin content id = " + contentlet.getIdentifier() +
									", skipping it and going to next one.", e);
						}
					}
					offSet += cicleValue;
					contentlets = conAPI.search("+working:true -structureName:Host", cicleValue, offSet, "", APILocator.getUserAPI().getSystemUser(), false);
					size = contentlets.size();
					Logger.info(this, "Processing next set of " + size);
				}
				Logger.info(this, "Modified "+ contentletsCounter + " Contentlets" );


				//Links
				Logger.info(this, "Searching for Links");
				int linksCounter = 0;
				List links = LinkFactory.getActiveLinks();

				Logger.info(this, "Found "+ links.size()+" Links");
				Iterator it = links.iterator();
				while (it.hasNext()) {
					Link link = (Link)it.next();
					//Folder parent = folderAPI.find(link.getParent(), systemUser, true);
					Logger.debug(this, "Modifying Link with id " + link.getIdentifier());
					String url = link.getUrl();
					matcher.reset(url);
					if(matcher.find()){
						if(url.endsWith(".dot")){
							url = url.substring(0, url.lastIndexOf(".dot"));
							url = url + replaceString;
						}else if(url.contains(".dot#") ){
							url = url.replace(".dot#", "." + replaceString + "#");
						}
						else if(url.contains(".dot?") ){
							url = url.replace(".dot?", "." + replaceString + "?");
						}
						//url = url.replaceAll(searchString, replaceString);
						link.setUrl(url);
						//InodeFactory.saveInode(link);
						HibernateUtil.saveOrUpdate(link);
						Identifier identifier = APILocator.getIdentifierAPI().find(link);
						WebAssetFactory.saveAsset(link, identifier);
						//WebAssetFactory.createAsset(link, systemUser.getUserId(), parent, identifier, false);
						linksCounter++;
					}
				}
				Logger.info(this, "Modified "+ linksCounter + " Links" );

				//Virtual Links
				Logger.info(this, "Searching for Virtual Links");
				int virtuallinksCounter = 0;
				List virtualLinks = VirtualLinkFactory.getVirtualLinks();
				Logger.info(this, "Found "+ links.size()+" Links");
				it = virtualLinks.iterator();
				while (it.hasNext()) {
					VirtualLink vlink = (VirtualLink)it.next();
					Logger.debug(this, "Modifying Virtual Link with inode " + vlink.getInode());
					String uri = vlink.getUri();
					matcher.reset(uri);
					if(matcher.find()){
						if(uri.endsWith(".dot")){
							uri = uri.substring(0, uri.lastIndexOf(".dot"));
							uri = uri + replaceString;
						}else if(uri.contains(".dot#") ){
							uri = uri.replace(".dot#", "." + replaceString + "#");
						}
						else if(uri.contains(".dot?") ){
							uri = uri.replace(".dot?", "." + replaceString + "?");
						}
						//uri = uri.replaceAll(searchString, replaceString);
						vlink.setUri(uri);
						//InodeFactory.saveInode(vlink);
						HibernateUtil.saveOrUpdate(vlink);
						VirtualLinksCache.removePathFromCache(vlink.getUrl());
					    VirtualLinksCache.addPathToCache(vlink);
						virtuallinksCounter++;
					}
				}
				Logger.info(this, "Modified "+ virtuallinksCounter + " Virtual Links" );

				//Containers
				Logger.info(this, "Searching for Containers");
				int containersCounter = 0;
				List containers = ContainerFactory.getActiveContainers();
				Logger.info(this, "Found "+ containers.size()+" Containers");
				it = containers.iterator();
				while (it.hasNext()) {
					Container container = (Container)it.next();
					Logger.debug(this, "Modifying Container with inode " + container.getInode());
					String code = container.getCode();
					boolean changed = false;
					matcher.reset(code);
					if(matcher.find()){
						code = code.replaceAll(searchString, replaceString);
						container.setCode(code);
						changed = true;
					}
					matcher.reset(code);
					if(matcher.find()){
						String preLoop = container.getPreLoop();
						preLoop = preLoop.replaceAll(searchString, replaceString);
						container.setPreLoop(preLoop);
						changed = true;
					}
					matcher.reset(code);
					if(matcher.find()){
						String postLoop = container.getPostLoop();
						postLoop = postLoop.replaceAll(searchString, replaceString);
						container.setPostLoop(postLoop);
						changed = true;
					}
					if(changed) {
						//InodeFactory.saveInode(container);
						Identifier identifier = APILocator.getIdentifierAPI().find(container);
						WebAssetFactory.createAsset(container, systemUser.getUserId(), identifier, false);
						container = (Container) WebAssetFactory.saveAsset(container,
								identifier);
						APILocator.getVersionableAPI().setWorking(container);
						containersCounter++;
					}
				}
				Logger.info(this, "Modified "+ containersCounter + " Containers" );

				//Templates
				Logger.info(this, "Searching for Templates");
				int templatesCounter = 0;
				List<Template> templates = APILocator.getTemplateAPI().findTemplates(systemUser, false, params, host.getIdentifier(), null, null,  "", 0, -1, "title");
				Logger.info(this, "Found "+ templates.size()+" Templates");
				for (Template template: templates) {
					Logger.debug(this, "Modifying Template with inode " + template.getInode());
					String body = template.getBody();
					matcher.reset(body);
					if(matcher.find()){
						body = body.replaceAll(searchString, replaceString);
						template.setBody(body);
						//InodeFactory.saveInode(template);
						APILocator.getTemplateAPI().saveTemplate(template, host, systemUser, false);
						APILocator.getVersionableAPI().setWorking(template);
						templatesCounter++;
					}
				}
				Logger.info(this, "Modified "+ templatesCounter + " Templates" );

				//HTML Pages
				Logger.info(this, "Searching for HTML Pages");
				int pagesCounter = 0;
				List<HTMLPage> pages = APILocator.getHTMLPageAPI().findHtmlPages(systemUser, false, params, host.getIdentifier(), null, null,  "", 0, -1, "title");
				Logger.info(this, "Found "+ pages.size()+" Pages");
				it = pages.iterator();
				for (HTMLPage htmlpage: pages) {
					Logger.debug(this, "Modifying HTML Page with inode " + htmlpage.getInode());
					String redirect = htmlpage.getRedirect();
					if(UtilMethods.isSet(redirect)){
						matcher.reset(redirect);
						if(matcher.find()){
							redirect = redirect.replaceAll(searchString, replaceString);
							htmlpage.setRedirect(redirect);
							//InodeFactory.saveInode(htmlpage);
							Folder parentFolder = folderAPI.find(htmlpage.getParent(),systemUser,true);
							Identifier identifier=APILocator.getIdentifierAPI().find(htmlpage);
							WebAssetFactory.createAsset(htmlpage, systemUser.getUserId(),
									parentFolder, identifier, false);
							HibernateUtil.flush();

							LiveCache.removeAssetFromCache(htmlpage);
							htmlpage = (HTMLPage) WebAssetFactory.saveAsset(htmlpage,
									identifier);
							APILocator.getVersionableAPI().setWorking(htmlpage);
							pagesCounter++;
						}
					}
				}
				Logger.info(this, "Modified "+ pagesCounter + " HTML Pages" );

				//Structure Widget
				Logger.info(this, "Searching for Structure Widgets Pages");
				int widgetsCounter = 0;
				List<Structure> structures = StructureFactory.getStructures("structuretype=2", "name", 0, 0, "asc");
				for(Structure st: structures){
					Field field = st.getFieldVar("widgetCode");
					Logger.debug(this, "Modifying Structure widget with inode " + st.getInode());
					String values = field.getValues();
					if(UtilMethods.isSet(values)){
						matcher.reset(values);
						if(matcher.find()){
							values = values.replaceAll(searchString, replaceString);
							field.setValues(values);
							//InodeFactory.saveInode(field);
							FieldFactory.saveField(field);
							widgetsCounter++;
						}
					}
				}
				Logger.info(this, "Modified "+ pagesCounter + " Structure Widgets" );

				HibernateUtil.commitTransaction();

				Logger.info(this, "Modifications Summary");
				Logger.info(this, contentletsCounter + " Contentlets modified!!");
				Logger.info(this, contentletsProccesed + " Contentlets proccesed!!");
				Logger.info(this, linksCounter + " Links modified!!");
				Logger.info(this, virtuallinksCounter + " Virtual Links modified!!" );
				Logger.info(this, containersCounter + " Containers modified!!");
				Logger.info(this, templatesCounter + " Templates modified!!");
				Logger.info(this, pagesCounter + " HTMLPages modified!!");
				Logger.info(this, widgetsCounter + " Structure Widgets modified!!");

				_deleteFiles(com.dotmarketing.util.WebKeys.Cache.CACHE_LIVE_FILES);
				_deleteFiles(com.dotmarketing.util.WebKeys.Cache.CACHE_WORKING_FILES);
				_flush(com.dotmarketing.util.WebKeys.Cache.CACHE_ALL_CACHES);
			}catch(Exception e){
				Logger.error(this, "A error occured running the replace, the transaction was rollbacked!! Exception message: " + e.getMessage());
				try {
					HibernateUtil.rollbackTransaction();
				} catch (DotHibernateException e1) {
					Logger.warn(this, "A error occured running the replace, the transaction was rollbacked: "+e1.getMessage());
					Logger.error(this,"unable to rollback transaction : " + e1.getMessage(),e1);
				}
				Logger.error(this,e.getMessage(),e);
			}

		}
	}

	/**
	 * Clean file cache
	 * @param cacheName
	 * @throws Exception
	 */
	private void _deleteFiles(String cacheName)
	throws Exception
	{
		try
		{
			String realPath = "";

			String velocityRootPath = Config.getStringProperty("VELOCITY_ROOT");
			if (velocityRootPath.startsWith("/WEB-INF")) {
				velocityRootPath = Config.CONTEXT.getRealPath(velocityRootPath);
			}

			if (cacheName.equals(com.dotmarketing.util.WebKeys.Cache.CACHE_LIVE_FILES))
			{
				realPath = velocityRootPath + File.separator + "live";
			}
			else
			{
				realPath = velocityRootPath + File.separator + "working";
			}
			File file = new File(realPath);
			if (file.isDirectory())
			{
				_deleteRoot(file);
				Logger.debug(ViewCMSMaintenanceAction.class,"The directory " + realPath + " has been deleted");
			}
		}
		catch(Exception ex)
		{
			Logger.error(ViewCMSMaintenanceAction.class,ex.toString());
			throw ex;
		}
	}

	/**
	 *
	 * @param root
	 * @return
	 * @throws Exception
	 */
	private boolean _deleteRoot(File root) throws Exception
	{
		boolean returnValue = true;
		File[] childs = root.listFiles();
		for(int i = 0; i < childs.length; i++)
		{
			File child = childs[i];
			if (child.isFile())
			{
				returnValue = returnValue && child.delete();
			}
			if (child.isDirectory())
			{
				returnValue = returnValue && _deleteRoot(child);
			}
		}
		return returnValue;
	}

	/**
	 * Flush the specified cache
	 * @param cacheName
	 * @throws Exception
	 */
	private void _flush(String cacheName)throws Exception{
		MaintenanceUtil.flushCache();
	}

}